# pylint: disable=unused-import,redefined-outer-name
from typing import TYPE_CHECKING

import pytest

from .azure import azurite  # noqa: F401
from .factory import TempUPathFactory
from .gcs import fake_gcs_server  # noqa: F401
from .s3 import MockedS3Server, s3_fake_creds_file, s3_server  # noqa: F401
from .utils import docker_client, monkeypatch_session  # noqa: F401

if TYPE_CHECKING:
    from pytest import FixtureRequest


@pytest.fixture
def aws_region_name():
    return "eu-west-1"


@pytest.fixture(scope="session")
def tmp_upath_factory(request: "FixtureRequest"):
    """Return a TempUPathFactory instance for the test session."""
    yield TempUPathFactory.from_request(request)


@pytest.fixture
def s3_path(tmp_upath_factory):
    """Temporary path on a mocked S3 remote."""
    yield tmp_upath_factory.mktemp("s3")


@pytest.fixture
def local_path(tmp_upath_factory, monkeypatch):
    """Return a temporary path."""
    ret = tmp_upath_factory.mktemp()
    monkeypatch.chdir(ret)
    yield ret


@pytest.fixture
def azure_path(tmp_upath_factory):
    """Return a temporary path."""
    yield tmp_upath_factory.mktemp("azure")


@pytest.fixture
def memory_path(tmp_upath_factory):
    """Return a temporary path in a MemoryFileSystem."""
    yield tmp_upath_factory.mktemp("memory")


@pytest.fixture
def gcs_path(tmp_upath_factory):
    """Return a temporary path."""
    yield tmp_upath_factory.mktemp("gcs")


@pytest.fixture
def tmp_upath(
    request: "FixtureRequest",
    tmp_upath_factory,
):
    """Temporary directory on different filesystems.

    Usage:
    >>> @pytest.mark.parametrize("tmp_upath", ["local", "s3", "azure"], indirect=True]) # noqa: E501
    >>> def test_something(tmp_upath):
    >>>     pass
    """
    param = getattr(request, "param", "local")
    if param == "local":
        return tmp_upath_factory.mktemp()
    elif param == "memory":
        return tmp_upath_factory.mktemp("memory")
    elif param == "s3":
        return tmp_upath_factory.mktemp("s3")
    elif param == "azure":
        return tmp_upath_factory.mktemp("azure")
    elif param == "gcs":
        return tmp_upath_factory.mktemp("gcs")
    raise ValueError(f"unknown {param=}")



def test_memory_fs_clean(tmp_upath_factory):
    mempath1 = tmp_upath_factory.mktemp("memory")
    mempath2 = tmp_upath_factory.mktemp("memory")

    foo = mempath1 / "foo"
    foo.write_text("foo")
    bar = mempath2 / "bar"
    bar.write_text("bar")

    assert (mempath1 / "foo").exists()
    assert (mempath2 / "bar").exists()
    assert not (mempath1 / "bar").exists()
    assert not (mempath2 / "foo").exists()



import os
from pathlib import Path

from pytest_servers.utils import is_pytest_session


def test_is_pytest_session():
    assert is_pytest_session()


def test_s3_fake_creds_file(
    s3_fake_creds_file,  # pylint: disable=unused-argument
):
    assert os.getenv("AWS_PROFILE") is None
    assert os.getenv("AWS_ACCESS_KEY_ID") == "pytest-servers"
    assert os.getenv("AWS_SECRET_ACCESS_KEY") == "pytest-servers"
    assert os.getenv("AWS_SECURITY_TOKEN") == "pytest-servers"
    assert os.getenv("AWS_SESSION_TOKEN") == "pytest-servers"
    assert (Path("~").expanduser() / ".aws").exists()



class PytestServersException(Exception):
    """Base class for all pytest-servers exceptions."""

    def __init__(self, msg: str, *args):
        assert msg
        self.msg = msg
        super().__init__(msg, *args)


class RemoteUnavailable(PytestServersException):
    """Raise when the given remote is not available"""

    def __init__(self, remote: str, *args):
        super().__init__(f"{remote} remote is not available", *args)



import tempfile
from typing import TYPE_CHECKING, Optional

from pytest import TempPathFactory
from upath import UPath

from pytest_servers.exceptions import RemoteUnavailable
from pytest_servers.local import LocalPath
from pytest_servers.utils import random_string, skip_or_raise_on

if TYPE_CHECKING:
    from pytest import FixtureRequest


class TempUPathFactory:
    """Factory for temporary directories with universal-pathlib and mocked servers"""  # noqa: E501

    mock_remotes = {
        "azure": ("azurite", "_azure_connection_string"),
        "gcs": ("fake_gcs_server", "_gcs_endpoint_url"),
        "s3": ("s3_server", "_s3_endpoint_url"),
    }

    def __init__(
        self,
        s3_endpoint_url: Optional[str] = None,
        azure_connection_string: Optional[str] = None,
        gcs_endpoint_url: Optional[str] = None,
    ):
        self._request: Optional["FixtureRequest"] = None

        self._local_path_factory: Optional["TempPathFactory"] = None
        self._azure_connection_string = azure_connection_string
        self._gcs_endpoint_url = gcs_endpoint_url
        self._s3_endpoint_url = s3_endpoint_url

    @classmethod
    def from_request(
        cls, request: "FixtureRequest", *args, **kwargs
    ) -> "TempUPathFactory":
        """Create a factory according to pytest configuration."""
        tmp_upath_factory = cls(*args, **kwargs)
        tmp_upath_factory._local_path_factory = TempPathFactory.from_config(
            request.config, _ispytest=True
        )
        tmp_upath_factory._request = request

        return tmp_upath_factory

    def _mock_remote_setup(self, fs: "str") -> None:
        try:
            mock_remote_fixture, remote_config_name = self.mock_remotes[fs]
        except KeyError:
            raise RemoteUnavailable(f"No mock remote available for fs: {fs}")

        if getattr(self, remote_config_name):  # remote is already configured
            return

        assert self._request
        try:
            remote_config = self._request.getfixturevalue(mock_remote_fixture)
            assert remote_config
            setattr(self, remote_config_name, remote_config)
        except Exception as exc:
            raise RemoteUnavailable(fs) from exc

    @skip_or_raise_on(ImportError, RemoteUnavailable)
    def mktemp(
        self, fs: str = "local", mock: bool = True, **kwargs
    ) -> "UPath":
        """Create a new temporary directory managed by the factory.

        :param fs:
            Filesystem type, one of
            - local (default)
            - memory
            - s3
            - azure
            - gcs

        :param mock:
            Set to False to use real remotes

        :returns:
            :class:`upath.Upath` to the new directory.
        """
        if fs == "local":
            return self.local_temp_path()
        elif fs == "memory":
            return self.memory_temp_path(**kwargs)

        if mock:
            self._mock_remote_setup(fs)

        if fs == "s3":
            return self.s3_temp_path(
                region_name="eu-south-1",
                endpoint_url=self._s3_endpoint_url,
                **kwargs,
            )
        elif fs == "azure":
            if not self._azure_connection_string:
                raise RemoteUnavailable("missing connection string")
            return self.azure_temp_path(
                connection_string=self._azure_connection_string, **kwargs
            )
        elif fs == "gcs":
            return self.gcs_temp_path(
                endpoint_url=self._gcs_endpoint_url,
                **kwargs,
            )
        else:
            raise ValueError(fs)

    def local_temp_path(self):
        mktemp = (
            self._local_path_factory.mktemp
            if self._local_path_factory is not None
            else tempfile.mktemp
        )
        return LocalPath(mktemp("pytest-servers"))

    def s3_temp_path(
        self, region_name: str, endpoint_url: Optional[str] = None, **kwargs
    ) -> UPath:
        """Creates a new S3 bucket and returns an UPath instance  .

        `endpoint_url` can be used to use custom servers (e.g. moto s3)."""
        client_kwargs = {}
        if endpoint_url:
            client_kwargs["endpoint_url"] = endpoint_url
        if region_name:
            client_kwargs["region_name"] = region_name

        bucket_name = f"pytest-servers-{random_string()}"
        path = UPath(
            f"s3://{bucket_name}", client_kwargs=client_kwargs, **kwargs
        )
        path.mkdir()
        return path

    def azure_temp_path(self, connection_string: str, **kwargs) -> UPath:
        """Creates a new container and returns an UPath instance"""
        container_name = f"pytest-servers-{random_string()}"
        path = UPath(
            f"az://{container_name}",
            connection_string=connection_string,
            **kwargs,
        )
        path.mkdir()
        return path

    def memory_temp_path(self, **kwargs) -> UPath:
        """Creates a new temporary in-memory path returns an UPath instance"""
        path = UPath(
            f"memory:/{random_string()}",
            **kwargs,
        )
        path.mkdir()
        return path

    def gcs_temp_path(
        self, endpoint_url: Optional[str] = None, **kwargs
    ) -> UPath:
        """Creates a new gcs bucket and returns an UPath instance.

        `endpoint_url` can be used to use custom servers
        (e.g. fake-gcs-server).
        """
        client_kwargs = {}
        if endpoint_url:
            client_kwargs["endpoint_url"] = endpoint_url

        bucket_name = f"pytest-servers-{random_string()}"
        path = UPath(f"gcs://{bucket_name}", **client_kwargs, **kwargs)
        path.mkdir()
        return path



import importlib
import sys
from typing import TYPE_CHECKING

import pytest
import upath.implementations.cloud
import upath.implementations.memory

from pytest_servers.local import LocalPath

if TYPE_CHECKING:
    from _pytest.mark import MarkDecorator

for module in ["s3fs", "adlfs", "gcsfs"]:
    try:
        importlib.import_module(module)
    except ModuleNotFoundError:
        pass


def skip_if_module_missing(name: str) -> "MarkDecorator":
    """Returns a mark that can be used to skip a test if a module is missing"""
    return pytest.mark.skipif(
        name not in sys.modules, reason=f"{name} is not installed"
    )


implementations = [
    pytest.param("local", LocalPath),
    pytest.param(
        "memory",
        upath.implementations.memory.MemoryPath,
    ),
    pytest.param(
        "s3",
        upath.implementations.cloud.S3Path,
        marks=[skip_if_module_missing("s3fs")],
    ),
    pytest.param(
        "azure",
        upath.implementations.cloud.AzurePath,
        marks=[skip_if_module_missing("adlfs")],
    ),
    pytest.param(
        "gcs",
        upath.implementations.cloud.GCSPath,
        marks=[skip_if_module_missing("gcsfs")],
    ),
]


@pytest.mark.parametrize(
    "fs,cls",
    implementations,
    ids=(param.values[0] for param in implementations),  # type: ignore
)
class TestTmpUPathFactory:
    def test_init(self, tmp_upath_factory, fs, cls):
        path = tmp_upath_factory.mktemp(fs)
        assert isinstance(path, cls)
        if fs != "gcs":  # for empty buckets on gcs path.exists() == False
            assert path.exists()

    def test_is_empty(self, tmp_upath_factory, fs, cls):
        path = tmp_upath_factory.mktemp(fs)
        assert list(path.iterdir()) == []

    def test_write_text(self, tmp_upath_factory, fs, cls):
        file = tmp_upath_factory.mktemp(fs) / "foo"
        assert file.write_text("bar")
        assert file.read_text() == "bar"

    def test_multiple_paths(self, tmp_upath_factory, fs, cls):
        path_1 = tmp_upath_factory.mktemp(fs)
        path_2 = tmp_upath_factory.mktemp(fs)
        assert str(path_1) != str(path_2)



import pathlib

from fsspec.implementations.local import LocalFileSystem


class LocalPath(type(pathlib.Path())):  # type: ignore[misc]
    fs = LocalFileSystem()

    @property
    def path(self):
        return str(self)



import logging
import os
import random
import socket
import string
import time

import pytest
from funcy import decorator

logger = logging.getLogger(__name__)


def wait_until(pred, timeout: float, pause: float = 0.1):
    start = time.perf_counter()
    while (time.perf_counter() - start) < timeout:
        value = pred()
        if value:
            return value
        time.sleep(pause)
    raise TimeoutError("timed out waiting")


def random_string(n: int = 6):
    return "".join(random.choices(string.ascii_lowercase, k=n))


@pytest.fixture(scope="session")
def monkeypatch_session():
    """Session-scoped monkeypatch"""
    from _pytest.monkeypatch import MonkeyPatch

    m = MonkeyPatch()
    yield m
    m.undo()


@pytest.fixture(scope="session")
def docker_client():
    """Run docker commands using the python API"""
    if os.environ.get("CI") and os.name == "nt":
        # docker-py currently fails to pull images on windows (?)
        yield None
        return

    try:
        import docker
    except ImportError:
        logger.warning("docker is not installed, run pip install docker")
        yield None
        return

    try:
        with pytest.deprecated_call():
            yield docker.from_env()
    except docker.errors.DockerException as exc:
        logging.exception(f"Failed to get docker client: {exc}")
        yield None


def is_pytest_session() -> bool:
    """returns true if currently running a pytest session"""
    return "PYTEST_CURRENT_TEST" in os.environ


@decorator
def skip_or_raise_on(call, *exceptions):
    """If in a pytest session, skips the current test when the given exceptions are raised"""  # noqa: E501
    try:
        return call()
    except exceptions as exc:
        if is_pytest_session():
            # just get the test name
            current_test = os.environ["PYTEST_CURRENT_TEST"].split()[0]
            pytest.skip(f"{current_test}: {exc}")
        raise


def get_free_port() -> None:
    retries = 3
    while retries >= 0:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.listen()
                free_port = s.getsockname()[1]
            return free_port
        except OSError as exc:
            retries -= 1
            if retries <= 0:
                raise SystemError("Could not get a free port") from exc



import logging
import time

import pytest
import requests

from .utils import get_free_port

logger = logging.getLogger(__name__)

GCS_DEFAULT_PORT = 4443


@pytest.fixture(scope="session")
def fake_gcs_server(docker_client):
    """Spins up a fake-gcs-server container. Returns the endpoint URL."""
    if docker_client is None:
        logger.warning(
            "fake-gcs-server cannot run because docker is not available."
        )
        yield None
        return

    # Some features, such as signed URLs and resumable uploads, require
    # `fake-gcs-server` to know the actual url it will be accessed
    # with. We can provide that with -public-host and -external-url.
    port = get_free_port()
    url = f"http://localhost:{port}"
    command = f"-scheme http -public-host {url} -external-url {url}"

    container = docker_client.containers.run(
        "fsouza/fake-gcs-server:latest",
        command=command,
        name="pytest-servers-fake-gcs-server",
        stdout=True,
        stderr=True,
        detach=True,
        remove=True,
        ports={f"{GCS_DEFAULT_PORT}/tcp": port},
    )
    retries = 3
    while True:
        try:
            r = requests.get(f"{url}/storage/v1/b", timeout=10)
            if r.ok:
                break
        except Exception as exc:  # noqa: E722 # pylint: disable=broad-except
            retries -= 1
            if retries < 0:
                raise SystemError from exc
            time.sleep(1)

    yield url

    container.stop()



import logging
import time

import pytest
import requests

AZURITE_PORT = 10000
AZURITE_URL = "http://localhost:{port}"
AZURITE_ACCOUNT_NAME = "devstoreaccount1"
AZURITE_KEY = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="  # noqa: E501
AZURITE_ENDPOINT = f"{AZURITE_URL}/{AZURITE_ACCOUNT_NAME}"
AZURITE_CONNECTION_STRING = f"DefaultEndpointsProtocol=http;AccountName={AZURITE_ACCOUNT_NAME};AccountKey={AZURITE_KEY};BlobEndpoint={AZURITE_ENDPOINT};"  # noqa: E501

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def azurite(docker_client):
    """Spins up an azurite container. Returns the connection string."""
    if docker_client is None:
        logger.warning(
            "Azurite cannot be started because docker is not available."
        )
        yield None
        return

    container = docker_client.containers.run(
        "mcr.microsoft.com/azure-storage/azurite",
        command="azurite-blob --loose --blobHost 0.0.0.0",
        name="pytest-servers-azurite",
        stdout=True,
        stderr=True,
        detach=True,
        remove=True,
        ports={f"{AZURITE_PORT}/tcp": None},  # assign a random port
    )

    port = docker_client.api.port("pytest-servers-azurite", AZURITE_PORT)[0][
        "HostPort"
    ]
    retries = 3
    while True:
        try:
            # wait until the container is up, even a 400 status code is ok
            r = requests.get(AZURITE_URL.format(port=port), timeout=10)
            if (
                r.status_code == 400
                and "Server" in r.headers
                and "Azurite" in r.headers["Server"]
            ):
                break
        except Exception as exc:  # noqa: E722 # pylint: disable=broad-except
            retries -= 1
            if retries < 0:
                raise SystemError from exc
            time.sleep(1)

    yield AZURITE_CONNECTION_STRING.format(port=port)

    container.stop()



import re
import shlex
import subprocess

import pytest
import requests
from funcy import silent

from .utils import wait_until


class MockedS3Server:
    def __init__(self):
        self.port = None
        self.proc = None

    @property
    def endpoint_url(self):
        if self.port is None:
            raise ValueError("start() must be called first.")
        return f"http://localhost:{self.port}"

    def __enter__(self):
        self.start()
        return self

    def start(self):
        """Starts moto s3 on a random port"""
        try:
            # should fail since we didn't start server yet
            r = requests.get(self.endpoint_url)
        except:  # noqa: E722, B001 # pylint: disable=bare-except
            pass
        else:
            if r.ok:
                raise RuntimeError("moto server already up")
        self.proc = subprocess.Popen(
            shlex.split(
                "moto_server s3 -p 0",  # get a random port
            ),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out = self.proc.stderr.readline()
        self.port = int(re.match(b".*http://127.0.0.1:(\\d+).*", out).group(1))
        wait_until(
            silent(lambda: requests.get(self.endpoint_url, timeout=5).ok), 5
        )

        return self

    def close(self):
        if self.proc is not None:
            self.proc.terminate()
            self.proc.wait()
        self.proc = None

    def __exit__(self, *exc_args):
        self.close()


@pytest.fixture(scope="session")
def s3_fake_creds_file(monkeypatch_session):
    # https://github.com/spulec/moto#other-caveats
    import pathlib

    aws_dir = pathlib.Path("~").expanduser() / ".aws"
    aws_dir.mkdir(exist_ok=True)

    aws_creds = aws_dir / "credentials"
    initially_exists = aws_creds.exists()

    if not initially_exists:
        aws_creds.touch()

    try:
        with monkeypatch_session.context() as m:
            try:
                m.delenv("AWS_PROFILE")
            except KeyError:
                pass
            m.setenv("AWS_ACCESS_KEY_ID", "pytest-servers")
            m.setenv("AWS_SECRET_ACCESS_KEY", "pytest-servers")
            m.setenv("AWS_SECURITY_TOKEN", "pytest-servers")
            m.setenv("AWS_SESSION_TOKEN", "pytest-servers")
            yield
    finally:
        if aws_creds.exists() and not initially_exists:
            aws_creds.unlink()


@pytest.fixture(scope="session")
def s3_server(
    s3_fake_creds_file,  # pylint: disable=unused-argument,redefined-outer-name
):
    """Spins up a moto s3 server. Returns the endpoint URL."""
    with MockedS3Server() as server:
        yield server.endpoint_url
